﻿namespace SpikingNeuronNetwork.Console
{
    using Lib;
    using Lib.NeuronModels;
    using System;
    using System.Collections.Generic;
    using System.Linq;

    /*! \mainpage Spiking Neuron Network Simulator Documentation
     *  \author Sam McKennoch
     *  \version 1.0
     *  \date 2007-2014
     *  \copyright MIT License
     *  \section Introduction
     *  The spiking neuron network simulator is designed to be a .NET C# based tool for simulating the and training networks of spiking neurons.  The tool source code described in this documentation is available below:
     *  <ul>
     *  <li><a href="thetaNeuronCode.zip">thetaNeuronCode.zip</a></li>
     *  </ul>
     *  This tool, though built to be expanded as a general spiking neuron network tool, primarily is focused on the theta neuron presently. Theta neurons are dynamic neuron models, that is to say their state is a function of time as well as a function of input spikes. Theta neurons represent an excellent trade-off between model complexity and analytical tractability. Other simpler spiking models such as the Linear Integrate and Fire model do not capture many important dynamic properties while other more complicated neuron models such as the Izhikevich or Hodgkins-Huxley Model have multiple state variables and are generally too complicated to enable the development of simple gradient-based learning equations. Much more information about theta neurons can be found <a href="NECO_2008_Draft.pdf">here</a>.  
<center>
<img src="NeuronSchematic.jpg" align="center" alt="Basic Theta Neuron Schematic" />
<p>
<b>Basic Theta Neuron Schematic</b>
</p>
</center>
\section S2 Theta Neuron Dynamics
<p>In order to understand the functionality of the theta neuron class it is helpful to have at least an overview understanding of the theta neuron dynamics. The trajectory of the state in a theta neuron is described by:</p>
<p>\f[ \frac{{d\theta }}{{dt}} = \left( {1 - \cos \theta } \right) + \alpha I\left( t \right)\left( {1 + \cos \theta } \right) \f]</p>
<p>where \f$\theta\f$ is the neuron phase (state); \f$\alpha\f$ is a scaling constant; and \f$I(t)\f$ is the input current that drives the dynamics. The input current is composed of the synaptic currents (generated by input spikes) and a baseline line current that determines the neuron dynamic properties. Thus the neuron state changes as a function of time, input spikes and neuron properties. It is important to note that in traditional neural networks used by engineers there is no concept of state, rather the output is instaneously generated from the input, making it difficult to accomdate time-based learning applications.</p>
<a class="anchor" id="sim_sec">
\section S3 Theta Neuron Simulation
<p>Many machine learning applications are ideal for theta neurons because of their simple yet dynamic nature. However, in order to simulate the theta neuron dynamic equation shown above, an inefficient process called numerical integration is typically used. The inefficiency of this process makes it difficult for theta neurons to be used in embedded applications such as robotics where the real-time response of the system is important. Thus, the theta neuron class also contains two ways to simulate theta neuron dynamics. Besides the numerical method, the Theta Neuon class contains event-driven simulation methods, which through some mathematical tricks only needs to perform calculations when an input spike is received rather than at every time iteration, thus enabling robotic applications such as this one <a href="IJCNN_2007.pdf">here</a>. </p>
\section S4 Future Work
<p>Many possible extensions are possible:</p>
        \todo Add Visualization and Web Interface (e.g. http://www.codeproject.com/Articles/5603/QuickGraph-A-100-C-graph-library-with-Graphviz-Sup)
        \todo Add Implementations for Training Algorithms RProp and QuickProp
        \todo Add Tests for rerun training after a reset and verfiy results are the same
        \todo Add Tests for Two consequtive trainings with different learning rates.  This test will involve making TrainingParameters tracked per epoch interval and specifying the  max number of epochs to train for (Cumulative) and the stopping criteria in the input
        \todo Add a Cosine type test
        \todo Add parallel computations where possible
        \todo Add constructors for multiple types of neurons in SpikingNeuronNetwork
        \todo Add neuron implementations for:
        \todo * QIF - http://hal.inria.fr/docs/00/10/35/00/PDF/eventdrivenQIF.pdf
        \todo * LIF - http://icwww.epfl.ch/~gerstner/SPNM/node26.html
        \todo * Izhikevich - http://www.abrg.group.shef.ac.uk/!DATA/attachment/0085.Humphries_simple_model_solutions_preprint.pdf, http://www.abrg.group.shef.ac.uk/code/othercode/simple_model_solutions_code.zip and http://faculty.washington.edu/shlizee/publications/rqif-det.pdf
        \todo * Fitz-Hugh Nagumo - http://eqworld.ipmnet.ru/en/solutions/npde/npde1103.pdf
        \todo Add Exponential synaptic current, see QIF ref, Bessel Functions
        \todo Add 2D models that can be solved analytically, perhaps is possible with dirac synaptic currents
     */

    /// <summary>
    /// Console Program Class
    /// </summary>
    public class Program
    {
        static void Main()
        {
	        const int numInputs = 4;
            const int maxNumOutputSpikes = 3;
            var inputSpikeTimes = new List<Spike>
                {
                    new Spike { Time = 2, NeuronIndex = 0}, 
                    new Spike { Time = 1, NeuronIndex = 2}, 
                    new Spike { Time = 1.5, NeuronIndex = 2}, 
                    new Spike { Time = 4, NeuronIndex = 3}, 
                };
	        
            // Construct theta neuron object with randomized weights
            var thetaNeuron = new ThetaNeuron(numInputs, 0.01);

	        // Display Neuron Properties
	        Console.WriteLine(thetaNeuron);

	        // Simulate Theta Neuron Numerically
	        thetaNeuron.Method = SimulationMethod.Numerical;
	        var outputSpikeTimes = thetaNeuron.RunThetaNeuron(inputSpikeTimes, maxNumOutputSpikes).ToList();

	        //Simulate Theta Neuron with Event-Driven Methodology
            thetaNeuron.ResetState();
	        thetaNeuron.Method = SimulationMethod.EventDriven;
            var outputSpikeTimes2 = thetaNeuron.RunThetaNeuron(inputSpikeTimes, maxNumOutputSpikes).ToList();

	        //Find Mean-Squared in Spike Times Between the Different Methods
            if (outputSpikeTimes.Count != outputSpikeTimes2.Count) 
            {
		        Console.WriteLine("The Two Simulation Methods Produced Different Numbers of Output Spikes!");
                Console.WriteLine();
            }
            else if (outputSpikeTimes.Count == 0) 
            {
		        Console.WriteLine("The Two Simulation Methods Produced No Output Spikes!");
                Console.WriteLine();
            }
	        else 
            {
                double meanSquaredError = 0;
                meanSquaredError += outputSpikeTimes.Select((t, j) => Math.Pow(t.Time - outputSpikeTimes2[j].Time, 2)).Sum();
                meanSquaredError /= outputSpikeTimes.Count;
                Console.WriteLine("Output Spike Mean-Squared Error Between Simulation Methods: ");
                Console.WriteLine("\t" + meanSquaredError);
                Console.WriteLine();
            }

	        //Stack two neurons sequentially
            if (outputSpikeTimes.Count > 0)
            {
		        Console.WriteLine("A second neuron is simulated in connection to the output of the previous neuron");
		        var thetaNeuron2 = new ThetaNeuron(1);
                var outputSpikeTimesList = outputSpikeTimes.Select((t, j) => new Spike{Time = t.Time, NeuronIndex = j}).ToList();
                thetaNeuron2.RunThetaNeuron(outputSpikeTimesList, maxNumOutputSpikes);
	        }

	        Console.ReadKey();
        }
    }
}
