﻿namespace SpikingNeuronNetwork.Lib.Interfaces
{
    using System.Collections.Generic;
    using Training;
    
    /// <summary>
    /// Spiking Neuron Interface
    /// </summary>
    public interface ISpikingNeuron
    {
        /// <summary>
        /// The neuron index within a network
        /// </summary>
        int NeuronIndex { get; set; }

        /// <summary>
        /// The network which the neuron is part of or null if neuron is stand alone
        /// </summary>
        SpikingNeuronNetwork NeuronNetwork { get; set; }
        
        /// <summary>
        /// Create verbose output
        /// </summary>
        bool Verbose { get; set; }

        /// <summary>
        /// The current neuron state
        /// </summary>
        NeuronState State { get; }
        
        /// <summary>
        /// The simulation method
        /// </summary>
        SimulationMethod Method { get; set; }

        /// <summary>
        /// The state to which a neuron returns to after spiking
        /// </summary>
        double PostSpikeState { get; }

        /// <summary>
        /// The initial state of the neuron
        /// </summary>
        double InitialState { get; }

        /// <summary>
        /// Indicates if the neuron is stand alone or part of a network
        /// </summary>
        bool IsStandAloneNeuron { get; }

        /// <summary>
        /// The remaining time until the neuron spikes absent further input spikes
        /// </summary>
        double RemainingTimeToSpike { get; }

        /// <summary>
        /// The next output spike time absent further input spikes
        /// </summary>
        double NextSpikeTime { get; }

        /// <summary>
        /// The number of inputs
        /// </summary>
        int NumInputs { get; }

        /// <summary>
        /// The weights associated with this neuron's inputs
        /// </summary>
        Dictionary<Synapse, double> Weights { get; }

        /// <summary>
        /// The Neuron Properties As A String
        /// </summary>
        string ToString();

        /// <summary>
        /// Clones a neuron by removing network association
        /// </summary>
        /// <returns>A stand alone neuron</returns>
        ISpikingNeuron Clone();

        /// <summary>
        /// Resets the neuron state to its initial conditions
        /// </summary>
        void ResetState();

        /// <summary>
        /// Runs the theta neuron, advancing the state and producing up to maxNumOutputSpikes 
        /// output spikes in response to inputSpikes
        /// </summary>
        /// <param name="inputSpikes">The input spikes</param>
        /// <param name="maxNumOutputSpikes">The maximum number of output spikes to be produced</param>
        /// <returns>The output spikes</returns>
        List<Spike> RunThetaNeuron(List<Spike> inputSpikes, int maxNumOutputSpikes);
 
        /// <summary>
        /// Starting from any state, process the neuron state forward in time 
        /// until the spike time and process the spike's effect.
        /// Note in the context of a network, if an output spike is produced on trajectory
        /// before the input spike is processed, the effect of the output spike could
        /// impact the neuron's state through feedback connections, so care should be taken
        /// </summary>
        /// <param name="spike">The input spike</param>
        /// <param name="spikeStats">Spike statistics output parameter</param>
        /// <returns>A collection of output spikes</returns>
        IEnumerable<Spike> ProcessSpike(Spike spike, out SpikeStats spikeStats);

        /// <summary>
        /// Calculates Post Spike Derivatives numerically, that is the derivate of the state after the current input spike
        /// relative to the derivative of the state after the previous input spike
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of post spike derivatives</returns>
        List<double> CalculatePostSpikeDerivativesNumerical(NeuronFiringHistory neuronFiringHistory);

        /// <summary>
        /// Calculates Post Spike Derivatives, that is the derivate of the state after the current input spike
        /// relative to the derivative of the state after the previous input spike
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of post spike derivatives</returns>
        List<double> CalculatePostSpikeDerivatives(NeuronFiringHistory neuronFiringHistory);

        /// <summary>
        /// Calculates Output Spike Derivatives numerically, that is the derivate of the output spike time relative to both the
        /// weight on each synapse and the input spike time on each synapse
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of output spike time derivatives</returns>
        Dictionary<Synapse, NeuronDerivativeParameters> CalculateOutputSpikeTimeDerivativesNumerical(NeuronFiringHistory neuronFiringHistory);

        /// <summary>
        /// Calculates Output Spike Derivatives, that is the derivate of the output spike time relative to both the
        /// weight on each synapse and the input spike time on each synapse
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of output spike time derivatives</returns>
        Dictionary<Synapse, NeuronDerivativeParameters> CalculateOutputSpikeTimeDerivatives(NeuronFiringHistory neuronFiringHistory);
    }
}
