﻿namespace SpikingNeuronNetwork.Lib.NeuronModels
{
    using Interfaces;
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Text;
    using Training;

    /// <summary>
    /// Spiking Neuron Abstract Class, Inherits From ISpikingNeuron
    /// </summary>
    public abstract class SpikingNeuron : ISpikingNeuron
    {
        /// <summary>
        /// Timestep
        /// </summary>
        public static double Timestep = 0.0001;
        
        /// <summary>
        /// Epsilon
        /// </summary>
        public static double Epsilon = 1e-10;
        
        /// <summary>
        /// Maximum number of output spikes to produce
        /// </summary>
        public const int MaxNumOutputSpikes = 5;

        /// <summary>
        /// Weights backing field
        /// </summary>
        protected Dictionary<Synapse, double> _weights;
        
        /// <summary>
        /// Spike Queue
        /// </summary>
        protected SpikePriorityQueue SpikeQueue;
        
        /// <summary>
        /// Random number generator
        /// </summary>
        private readonly Random _random = new Random();

        /// <summary>
        /// The neuron index within a network
        /// </summary>
        public int NeuronIndex { get; set; }

        /// <summary>
        /// The network which the neuron is part of or null if neuron is stand alone
        /// </summary>
        public SpikingNeuronNetwork NeuronNetwork { get; set; }

        /// <summary>
        /// Create verbose output
        /// </summary>
        public bool Verbose { get; set; }

        /// <summary>
        /// The current neuron state
        /// </summary>
        public NeuronState State { get; protected set; }

        /// <summary>
        /// The simulation method
        /// </summary>
        public SimulationMethod Method { get; set; }

        /// <summary>
        /// The state to which a neuron returns to after spiking
        /// </summary>
        public abstract double PostSpikeState { get; }

        /// <summary>
        /// The spike threshold, the state at which an output spike is produced
        /// </summary>
        public abstract double SpikeThreshold { get; }

        /// <summary>
        /// The initial state of the neuron
        /// </summary>
        public abstract double InitialState { get; }

        /// <summary>
        /// The remaining time until the neuron spikes absent further input spikes
        /// </summary>
        public abstract double RemainingTimeToSpike { get; }

        /// <summary>
        /// Maximum number of numerical integration iterations
        /// </summary>
        public abstract double MaxNumericalIntegrationIterations { get; }
        
        /// <summary>
        /// Operating region
        /// </summary>
        public abstract OperatingRegion OperatingRegion { get; }

        /// <summary>
        /// The next output spike time absent further input spikes
        /// </summary>
        public double NextSpikeTime
        {
            get { return RemainingTimeToSpike + State.Time; }
        }

        /// <summary>
        /// Indicates if the neuron is stand alone or part of a network
        /// </summary>
        public bool IsStandAloneNeuron
        {
            get { return NeuronNetwork == null; }
        }

        /// <summary>
        /// The number of inputs
        /// </summary>
        public int NumInputs
        {
            get
            {
                return IsStandAloneNeuron
                           ? Weights.Count(x => Math.Abs(x.Value) > Epsilon)
                           : Weights.Where(x => x.Key.InputNeuronIndex < NeuronNetwork.NumInputs)
                                    .Count(x => Math.Abs(x.Value) > Epsilon);
            }
        }

        /// <summary>
        /// The weights associated with this neuron's inputs
        /// </summary>
        public Dictionary<Synapse, double> Weights
        {
            get { return IsStandAloneNeuron ? _weights : NeuronNetwork.GetNeuronWeights(NeuronIndex); }
        }

        /// <summary>
        /// Creates a new instance of SpikingNeuron with no inputs
        /// </summary>
        protected SpikingNeuron()
            : this(0)
        {
        }

        /// <summary>
        /// Creates a new instance of SpikingNeuron an associates it to a network
        /// </summary>
        /// <param name="network">The neuron's network</param>
        /// <param name="neuronIndex">The neuron's index in the network</param>
        protected SpikingNeuron(SpikingNeuronNetwork network, int neuronIndex)
        {
            NeuronNetwork = network;
            NeuronIndex = neuronIndex;
            ResetState();
            Verbose = true;
            Method = SimulationMethod.EventDriven;
        }

        /// <summary>
        /// Creates a new instance of SpikingNeuron with given weights
        /// </summary>
        /// <param name="weights">The neuron's weights</param>
        protected SpikingNeuron(IReadOnlyList<double> weights)
        {
            var numInputs = weights.Count;
            NeuronIndex = numInputs;
            _weights = new Dictionary<Synapse, double>(weights.Count);
            for (var k = 0; k < numInputs; k++)
            {
                _weights.Add(new Synapse(k, NeuronIndex), weights[k]);
            }
            ResetState();
            Verbose = true;
            Method = SimulationMethod.Numerical;
        }

        /// <summary>
        /// Creates a new instance of SpikingNeuron with a given number of inputs and an initial weight
        /// </summary>
        /// <param name="numInputs">The number of inputs</param>
        /// <param name="weightIni">The initial weight</param>
        protected SpikingNeuron(int numInputs = 3, double weightIni = 1e6)
        {
            _weights = new Dictionary<Synapse, double>(numInputs);
            NeuronIndex = numInputs;
            for (var k = 0; k < numInputs; k++)
            {
                var newWeight = weightIni > 100 ? 2*_random.NextDouble() - 1 : weightIni;
                _weights.Add(new Synapse(k, NeuronIndex), newWeight);
            }
            ResetState();
            Verbose = true;
            Method = SimulationMethod.Numerical;
        }

        /// <summary>
        /// Clones a neuron by removing network association
        /// </summary>
        /// <returns>A stand alone neuron</returns>
        public abstract ISpikingNeuron Clone();

        /// <summary>
        /// Runs the theta neuron, advancing the state and producing up to maxNumOutputSpikes 
        /// output spikes in response to inputSpikes
        /// </summary>
        /// <param name="inputSpikes">The input spikes</param>
        /// <param name="maxNumOutputSpikes">The maximum number of output spikes to be produced</param>
        /// <returns>The output spikes</returns>
        public abstract List<Spike> RunThetaNeuron(List<Spike> inputSpikes, int maxNumOutputSpikes = MaxNumOutputSpikes);

        /// <summary>
        /// Updates the neuron's state variable, given a spike is currently being received on a synapse with weight
        /// </summary>
        /// <param name="weight">The synaptic weight</param>
        /// <returns>A collection of output spikes</returns>
        public abstract IEnumerable<Spike> UpdateStateVariableOnSpike(double weight);

        /// <summary>
        /// Advances the neuron's state to newTime
        /// </summary>
        /// <param name="newTime">The new neuron time</param>
        /// <returns>A collection of output spikes</returns>
        public abstract IEnumerable<Spike> AdvanceNeuronState(double newTime);

        /// <summary>
        /// Calculates Post Spike Derivatives, that is the derivate of the state after the current input spike
        /// relative to the derivative of the state after the previous input spike
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of post spike derivatives</returns>
        public abstract List<double> CalculatePostSpikeDerivatives(NeuronFiringHistory neuronFiringHistory);

        /// <summary>
        /// Calculates Output Spike Derivatives, that is the derivate of the output spike time relative to both the
        /// weight on each synapse and the input spike time on each synapse
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of output spike time derivatives</returns>
        public abstract Dictionary<Synapse, NeuronDerivativeParameters> CalculateOutputSpikeTimeDerivatives(NeuronFiringHistory neuronFiringHistory);
        
        /// <summary>
        /// Calculate the state derivative as a function of the state variable
        /// </summary>
        /// <param name="stateVariable">The state variable</param>
        /// <returns>The state derivative</returns>
        public abstract double StateDerivative(double stateVariable);

        /// <summary>
        /// Runs the theta neuron numerically, advancing the state and producing up to maxNumOutputSpikes 
        /// output spikes in response to spikes in the SpikeQueue
        /// </summary>
        /// <param name="maxNumOutputSpikes">The maximum number of output spikes to be produced</param>
        /// <returns>The output spikes</returns>
        public List<Spike> RunThetaNeuronNumerically(int maxNumOutputSpikes = MaxNumOutputSpikes)
        {
            var outputSpikes = new List<Spike>();
            if (SpikeQueue.Count == 0)
            {
                return outputSpikes;
            }
            var statePast = State.StateVariable;
            var allInputSpikeProcessed = false;
            var currentSpike = SpikeQueue.Dequeue();

            //The maximum number of numerical integration iterations
            var intLimit = MaxNumericalIntegrationIterations;

            var doIntegration = true;
            while (doIntegration)
            {
                // Forward Euler Numeric Integration
                State.StateVariable = statePast + Timestep * StateDerivative(statePast);

                if (!allInputSpikeProcessed && (currentSpike.Time >= State.Time) && (currentSpike.Time < (State.Time + Timestep)))
                {
                    var synapse = new Synapse(currentSpike.NeuronIndex, NeuronIndex);
                    var spikeStats = new SpikeStats();
                    try 
                    {
                        spikeStats = new SpikeStats
                            {
                                PreSpikeState = new NeuronState {StateVariable = State.StateVariable, Time = currentSpike.Time},
                                Weight = Weights[synapse],
                                Synapse = synapse
                            };
                    }
                    catch (KeyNotFoundException)
                    {
                        Console.WriteLine("Synapse: " + synapse);
                        Console.WriteLine("Weights: " + GetWeightsString(Weights));
                        throw;
                    }
                    UpdateStateVariableOnSpike(spikeStats.Weight);
                    spikeStats.PostSpikeState = new NeuronState { StateVariable = State.StateVariable, Time = currentSpike.Time };
                    if (Verbose)
                    {
                        Console.WriteLine("*** Spike From Neuron " + currentSpike.NeuronIndex + " To Neuron " + NeuronIndex + " ***");
                        Console.WriteLine(spikeStats);
                    }
                    if (SpikeQueue.Count > 0)
                    {
                        currentSpike = SpikeQueue.Dequeue();
                    }
                    else
                    {
                        allInputSpikeProcessed = true;
                    }
                }

                // If all input spikes have been processed and phase is quiescent
                // then no further output spikes are possible
                if (allInputSpikeProcessed && (OperatingRegion != OperatingRegion.SpikeGeneration && OperatingRegion != OperatingRegion.TonicSpiking))
                {
                    break;
                }

                // Check for output spike
                if (State.StateVariable > SpikeThreshold && statePast <= SpikeThreshold)
                {
                    outputSpikes.Add(new Spike { NeuronIndex = NeuronIndex, Time = State.Time });
                    if (Verbose)
                    {
                        Console.WriteLine("---> " + outputSpikes.Last() + "\n");
                    }
                    State.StateVariable = PostSpikeState;
                    if (outputSpikes.Count == maxNumOutputSpikes)
                    {
                        return outputSpikes;
                    }
                }

                // Special Case of Refractory Firing
                if (State.StateVariable > PostSpikeState && statePast <= PostSpikeState)
                {
                    State.StateVariable = SpikeThreshold - (State.StateVariable + PostSpikeState);
                }

                // Prepare for Next Iteration
                statePast = State.StateVariable;
                State.Time += Timestep;
                if (State.Time <= intLimit) continue;
                if (!(OperatingRegion == OperatingRegion.SpikeGeneration ||
                        (OperatingRegion == OperatingRegion.TonicSpiking && (outputSpikes.Count == 0 || currentSpike.Time > outputSpikes.Last().Time))))
                {
                    doIntegration = false;
                }
            }

            return outputSpikes;
        }

        /// <summary>
        /// Calculates Post Spike Derivatives numerically, that is the derivate of the state after the current input spike
        /// relative to the derivative of the state after the previous input spike
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of post spike derivatives</returns>
        public List<double> CalculatePostSpikeDerivativesNumerical(NeuronFiringHistory neuronFiringHistory)
        {
            // Calculate all DthpDthpm1 terms (Equation 3.1.4 in NECO Paper, Although MIMO Version is Used Here)
            var postSpikeDerivativeProducts = new List<double>();
            SpikeStats previousSpikeStat = null;
            var initialState = State;
            const double offset = 0.001;
            foreach (var spikeStat in neuronFiringHistory.SpikeStats)
            {
                if (previousSpikeStat == null)
                {
                    previousSpikeStat = spikeStat;
                    continue;
                }

                var currentSpike = new Spike
                {
                    NeuronIndex = spikeStat.Synapse.InputNeuronIndex,
                    Time = spikeStat.PreSpikeState.Time
                };
                
                State = new NeuronState { StateVariable = previousSpikeStat.PostSpikeState.StateVariable, Time = previousSpikeStat.PostSpikeState.Time };
                State.StateVariable -= offset;
                SpikeStats spikeStatsGradDelta1;
                ProcessSpike(currentSpike, out spikeStatsGradDelta1);
                State = new NeuronState { StateVariable = previousSpikeStat.PostSpikeState.StateVariable, Time = previousSpikeStat.PostSpikeState.Time };
                State.StateVariable += offset;
                SpikeStats spikeStatsGradDelta2;
                ProcessSpike(currentSpike, out spikeStatsGradDelta2);

                postSpikeDerivativeProducts.Add((spikeStatsGradDelta2.PostSpikeState.StateVariable - spikeStatsGradDelta1.PostSpikeState.StateVariable) / (2 * offset));
                previousSpikeStat = spikeStat;
            }

            State = initialState;
            return postSpikeDerivativeProducts;
        }

        /// <summary>
        /// Calculates Output Spike Derivatives numerically, that is the derivate of the output spike time relative to both the
        /// weight on each synapse and the input spike time on each synapse
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of output spike time derivatives</returns>
        public Dictionary<Synapse, NeuronDerivativeParameters> CalculateOutputSpikeTimeDerivativesNumerical(NeuronFiringHistory neuronFiringHistory)
        {
            // Convert neuronFiringHistory to input spikes
            var inputSpikes = neuronFiringHistory.SpikeStats.Select(spikeStat => new Spike
            {
                NeuronIndex = spikeStat.Synapse.InputNeuronIndex,
                Time = spikeStat.PreSpikeState.Time
            }).ToList();

            if (neuronFiringHistory.OutputSpikes.Count > 1)
            {
                throw new ArgumentException("MIMO Training Not Implemented");
            }

            if (neuronFiringHistory.OutputSpikes.Count == 0)
            {
                throw new ArgumentException("Actual Output Spike Times are Missing");
            }

            var inputSynapses = Weights.Select(x => x.Key).ToList();
            var outputSpikeTimeDerivatives = inputSynapses.ToDictionary(synapse => synapse, synapse => new NeuronDerivativeParameters
                {
                    OutputSpikeTimeToInputSpikeTimeDerivative = 0, OutputSpikeTimeToWeightDerivative = 0
                });

            const double offset = 0.00001;
            var initialState = State;
            foreach (var spikeStat in neuronFiringHistory.SpikeStats)
            {
                // Modify weight
                var weight = Weights.First(x => x.Key.InputNeuronIndex == spikeStat.Synapse.InputNeuronIndex);
                
                SetSynapticWeight(spikeStat.Synapse, weight.Value - offset);
                ResetState();
                var outputSpikeTime1 = RunThetaNeuron(inputSpikes).First().Time;
                
                SetSynapticWeight(spikeStat.Synapse, weight.Value + offset);
                ResetState();
                var outputSpikeTime2 = RunThetaNeuron(inputSpikes).First().Time;

                SetSynapticWeight(spikeStat.Synapse, weight.Value);
                var outputToWeightDerivative = (outputSpikeTime2 - outputSpikeTime1) / (2 * offset);

                // Modify input spike time
                var currentSpike = inputSpikes.First(x => x.NeuronIndex == spikeStat.Synapse.InputNeuronIndex);

                currentSpike.Time -= offset;
                ResetState();
                var outputSpikeTime3 = RunThetaNeuron(inputSpikes).First().Time;

                currentSpike.Time += 2*offset;
                ResetState();
                var outputSpikeTime4 = RunThetaNeuron(inputSpikes).First().Time;

                var outputToInputSpikeTimeDerivative = (outputSpikeTime4 - outputSpikeTime3) / (2 * offset);

                var derivativeParameters = new NeuronDerivativeParameters
                    {
                        OutputSpikeTimeToWeightDerivative = outputToWeightDerivative,
                        OutputSpikeTimeToInputSpikeTimeDerivative = outputToInputSpikeTimeDerivative
                    };
                outputSpikeTimeDerivatives[spikeStat.Synapse] = derivativeParameters;
            }

            State = initialState;
            return outputSpikeTimeDerivatives;
        }

        /// <summary>
        /// Starting from any state, process the neuron state forward in time 
        /// until the input spike time and process the spike's effect.
        /// Note in the context of a network, if an output spike is produced on trajectory
        /// before the input spike is processed, the effect of the output spike could
        /// impact the neuron's state through feedback connections, so care should be taken
        /// </summary>
        /// <param name="spike">The input spike</param>
        /// <param name="spikeStats">Spike statistics output parameter</param>
        /// <returns>A collection of output spikes</returns>
        public IEnumerable<Spike> ProcessSpike(Spike spike, out SpikeStats spikeStats)
        {
            // Protect against non-causality
            if (State.Time > spike.Time)
            {
                spikeStats = null;
                return new List<Spike>();
            }

            var outputSpikeTimes = AdvanceNeuronState(spike.Time).ToList();
            spikeStats = new SpikeStats
            {
                PreSpikeState = new NeuronState
                {
                    StateVariable = State.StateVariable,
                    Time = State.Time
                },
                Weight = Weights.First(x => x.Key.InputNeuronIndex == spike.NeuronIndex).Value,
                Synapse = new Synapse(spike.NeuronIndex, NeuronIndex)
            };

            var updateSpikes = UpdateStateVariableOnSpike(spikeStats.Weight).ToList();
            var firedOnSpike = updateSpikes.Count > 0;
            outputSpikeTimes.AddRange(updateSpikes);

            spikeStats.PostSpikeState = new NeuronState
            {
                StateVariable = State.StateVariable,
                Time = State.Time
            };

            if (Verbose)
            {
                Console.WriteLine("*** Spike From Neuron " + spike.NeuronIndex + " To Neuron " + NeuronIndex + " ***");
                Console.WriteLine(spikeStats);
                if (outputSpikeTimes.Count > 0)
                {
                    foreach (var outputSpikeTime in outputSpikeTimes)
                    {
                        Console.WriteLine("---> " + outputSpikeTime);
                    }
                }
                else
                {
                    Console.WriteLine("---> No Output Spikes Produced");
                }

                if (firedOnSpike)
                {
                    Console.WriteLine("---> At least one output spike fired directly from input spike");
                }

                Console.WriteLine();
            }

            return outputSpikeTimes;
        }

        /// <summary>
        /// Resets the neuron state to its initial conditions
        /// </summary>
        public void ResetState()
        {
            State = new NeuronState
                {
                    StateVariable = InitialState,
                    Time = 0
                };
        }

        /// <summary>
        /// Sets synaptic weight to newValue
        /// </summary>
        /// <param name="synapse">The synapse</param>
        /// <param name="newValue">The new value</param>
        public void SetSynapticWeight(Synapse synapse, double newValue)
        {
            if (IsStandAloneNeuron)
            {
                if (_weights.ContainsKey(synapse))
                {
                    _weights[synapse] = newValue;
                }
                else
                {
                    _weights.Add(synapse, newValue);    
                }                
            }
            else
            {
                NeuronNetwork.SetSynapticWeight(synapse, newValue);
            }
        }

        /// <summary>
        /// Get a string representation of the weights dictionary
        /// </summary>
        /// <param name="weights">The weights dictionary</param>
        /// <returns>A string representation of the weights dictionary</returns>
        protected static string GetWeightsString(Dictionary<Synapse, double> weights)
        {
            var weightsStringBuilder = new StringBuilder();
            weights.Keys.ToList().ForEach(x => weightsStringBuilder.AppendLine("\t" + weights[x] + "(" + x.InputNeuronIndex + "->" + x.OutputNeuronIndex + ")"));
            return weightsStringBuilder.ToString();
        }        
    }
}
