﻿namespace SpikingNeuronNetwork.Lib
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;

    /// <summary>
    /// Spike Priority Queue used for determining what spike to process next based on spike timing
    /// </summary>
    public class SpikePriorityQueue : ICollection<Spike>
    {
        /// <summary>
        /// Sorted dictionary used for background queueing of spikes
        /// </summary>
        private readonly SortedDictionary<double, Queue<int>> _baseQueue;

        /// <summary>
        /// Creates a new spike priority queue
        /// </summary>
        /// <param name="inputSpikes">Collection of input spikes</param>
        /// <param name="currentTime">The current time</param>
        /// <returns>A new spike priority queue</returns>
        public static SpikePriorityQueue CreateSpikeQueue(IEnumerable<Spike> inputSpikes, double currentTime)
        {
            var inputSpikeTimesList = inputSpikes.ToList();
            var spikeQueue = new SpikePriorityQueue(inputSpikeTimesList);
            if (spikeQueue.Count == 0)
            {
                return spikeQueue;
            }

            // Protect against non-causality
            while (currentTime > spikeQueue.PeekTime())
            {
                spikeQueue.Dequeue();
            }

            return spikeQueue;
        }


        /// <summary>
        /// Initializes a new instance of priority queue with specified input spikes
        /// </summary>
        /// <param name="inputSpikes">Collection of input spikes to add to the queue on creation</param>
        public SpikePriorityQueue(IEnumerable<Spike> inputSpikes)
            : this()
        {
            foreach (var inputSpike in inputSpikes)
            {
                Add(inputSpike);
            }
        }

        /// <summary>
        /// Initializes a new instance of the spike priority queue
        /// </summary>
        public SpikePriorityQueue()
        {
            _baseQueue = new SortedDictionary<double, Queue<int>>();
        }

        #region Priority queue operations

        /// <summary>
        /// Enqueues spike into the priority queue
        /// </summary>
        public void Enqueue(Spike inputSpike)
        {
            if (_baseQueue.ContainsKey(inputSpike.Time))
            {
                _baseQueue[inputSpike.Time].Enqueue(inputSpike.NeuronIndex);
            }
            else
            {
                _baseQueue.Add(inputSpike.Time, new Queue<int>(new[] {inputSpike.NeuronIndex}));    
            }            
        }

        /// <summary>
        /// Dequeues spike with next firing time and returns the spike/> 
        /// </summary>
        /// <returns>The dequeued spike</returns>
        /// <remarks>
        /// Method throws <see cref="InvalidOperationException"/> if priority queue is empty
        /// </remarks>
        public Spike Dequeue()
        {
            if (!IsEmpty)
            {
                var result = Peek();
                Remove(result);
                return result;
            }
            throw new InvalidOperationException("Priority queue is empty");
        }

        /// <summary>
        /// Dequeues spike with next firing time and returns its index
        /// </summary>
        /// <returns>index of the dequeued neuron</returns>
        /// <remarks>
        /// Method throws <see cref="InvalidOperationException"/> if priority queue is empty
        /// </remarks>
        public int DequeueValue()
        {
            return Dequeue().NeuronIndex;
        }

        /// <summary>
        /// Returns the spike with next firing time, without removing it from the queue
        /// </summary>
        /// <returns>The spike with next firing time</returns>
        /// <remarks>
        /// Method throws <see cref="InvalidOperationException"/> if priority queue is empty
        /// </remarks>
        public Spike Peek()
        {
            if (!IsEmpty)
            {
                var first = _baseQueue.First();
                return new Spike { Time = first.Key, NeuronIndex = first.Value.Peek()};
            }
            throw new InvalidOperationException("Priority queue is empty");
        }

        /// <summary>
        /// Returns the next firing time of a spike, without removing the spike from the queue
        /// </summary>
        /// <returns>The next spike firing time</returns>
        /// <remarks>
        /// Method throws <see cref="InvalidOperationException"/> if priority queue is empty
        /// </remarks>
        public double PeekTime()
        {
            return Peek().Time;
        }

        /// <summary>
        /// Gets whether priority queue is empty
        /// </summary>
        public bool IsEmpty
        {
            get { return _baseQueue.Count == 0; }
        }

        /// <summary>
        /// Updates a spike in the queue with a new firing time
        /// </summary>
        /// <param name="oldSpike">The old spike to be updated</param>
        /// <param name="newFiringTime">The new spike firing time</param>
        /// <returns></returns>
        public bool Update(Spike oldSpike, double newFiringTime)
        {
            if (!Remove(oldSpike))
                return false;
            if (!Double.IsInfinity(newFiringTime))
            {
                Add(new Spike { Time = newFiringTime, NeuronIndex = oldSpike.NeuronIndex });
            }
            return true;
        }

        /// <summary>
        /// Adds a range of spikes to the priority queue
        /// </summary>
        /// <param name="spikes">The collection of spikes to add</param>
        public void AddRange(ICollection<Spike> spikes)
        {
            foreach (var spike in spikes)
            {
                Add(spike);
            }
        }

        #endregion

        #region ICollection<Spike> implementation

        /// <summary>
        /// Enqueus spike into priority queue
        /// </summary>
        /// <param name="spike">spike to add</param>
        public void Add(Spike spike)
        {
            Enqueue(spike);
        }

        /// <summary>
        /// Clears the collection
        /// </summary>
        public void Clear()
        {
            _baseQueue.Clear();
        }

        /// <summary>
        /// Determines whether the priority queue contains a specific spike
        /// </summary>
        /// <param name="spike">The spike to locate in the priority queue</param>
        /// <returns><c>true</c> if spike is found in the priority queue; otherwise, <c>false.</c> </returns>
        public bool Contains(Spike spike)
        {
            if (!_baseQueue.ContainsKey(spike.Time))
            {
                return false;
            }
            return _baseQueue[spike.Time].Contains(spike.NeuronIndex);
        }

        /// <summary>
        /// Gets number of spikes in the priority queue
        /// </summary>
        public int Count
        {
            get { return _baseQueue.Sum(x => x.Value.Count); }
        }

        /// <summary>
        /// Copies the spikes of the priority queue to an Array, starting at a particular Array index. 
        /// </summary>
        /// <param name="array">The one-dimensional Array that is the destination of the elements copied from the priority queue. The Array must have zero-based indexing. </param>
        /// <param name="arrayIndex">The zero-based index in array at which copying begins.</param>
        /// <remarks>
        /// It is not guaranteed that items will be copied in the sorted order.
        /// </remarks>
        public void CopyTo(Spike[] array, int arrayIndex)
        {
            var arrayList = new List<Spike>();
            foreach (var kvp in _baseQueue)
            {
                var queueList = kvp.Value.ToList();
                arrayList.AddRange(queueList.Select(item => new Spike { Time = kvp.Key, NeuronIndex = item}));
            }
            arrayList.CopyTo(array, arrayIndex);
        }

        /// <summary>
        /// Gets a value indicating whether the collection is read-only. 
        /// </summary>
        /// <remarks>
        /// For priority queue this property returns <c>false</c>.
        /// </remarks>
        public bool IsReadOnly
        {
            get { return false; }
        }

        /// <summary>
        /// Removes the first occurrence of a specific spike from the priority queue. 
        /// </summary>
        /// <param name="spike">The spike to remove from the ICollection of spikes. </param>
        /// <returns><c>true</c> if item was successfully removed from the priority queue.
        /// This method returns false if item is not found in the collection. </returns>
        public bool Remove(Spike spike)
        {
            // find element in the collection and remove it
            if (!_baseQueue.ContainsKey(spike.Time))
            {
                return false;
            }
            if (_baseQueue[spike.Time].Count == 1)
            {
                _baseQueue.Remove(spike.Time);
                return true;
            }
            // This is expensive, should be avoided
            var queueList = _baseQueue[spike.Time].ToList();
            queueList.Remove(spike.NeuronIndex);
            _baseQueue[spike.Time] = new Queue<int>(queueList);
            return true;
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>Enumerator</returns>
        /// <remarks>
        /// Returned enumerator does not iterate elements in sorted order.</remarks>
        public IEnumerator<Spike> GetEnumerator()
        {
            return (from kvp in _baseQueue
                    from item in kvp.Value.ToList()
                    select new Spike{ Time = kvp.Key, NeuronIndex = item}).GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>Enumerator</returns>
        /// <remarks>
        /// Returned enumerator does not iterate elements in sorted order.</remarks>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion        
    }
}
