﻿namespace SpikingNeuronNetwork.Lib
{
    using Interfaces;
    using NeuronModels;
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Training;

    /// <summary>
    /// Spiking Neuron Network Class
    /// </summary>
    public class SpikingNeuronNetwork
    {
        /// <summary>
        /// Gets or sets the network id
        /// </summary>
        public int Id { get; set; }

        /// <summary>
        /// Gets the number of neurons per layer for layered networks
        /// </summary>
        public IList<int> NumNeuronsPerLayer { get; private set; }

        /// <summary>
        /// Gets the number of inputs connections to the network
        /// </summary>
        public int NumInputs
        {
            get { return _weightMatrix.Cols - NumNeuronsPerLayer.Sum(); }
        }

        /// <summary>
        /// Indicates if the network is layered
        /// </summary>
        public bool IsLayered
        {
            get { return _isLayered; }
        }

        private Dictionary<int, SpikingNeuron> _neurons;
        private const double Epsilon = 1e-10;
        private readonly Matrix _weightMatrix; //Should supercede the connections in the neuron objects
        private readonly Random _random = new Random();
        private readonly bool _isLayered;
       
        /// <summary>
        /// Creates a Spiking Neuron Network Object With Random Connectivity Topology
        /// </summary>
        /// <param name="numInputs">Number of inputs to the network</param>
        /// <param name="numNeurons">Number of neurons within the network</param>
        /// <param name="neuronType">Neuron Type, inherited from ISpikingNeuron</param>
        /// <param name="weightIni">The initial weight between neurons.  If weightIni is > 100, then random weights are used with uniform distribution between -1 and 1, with non-zero probability of 0.5</param>
        /// <param name="neuronProperties">Optional neuron properties to assign to each neuron</param>
        public SpikingNeuronNetwork(int numInputs, int numNeurons, Type neuronType, double weightIni = 1e6, Dictionary<string, string> neuronProperties = null)
        {
            // Create Network with Random Topology and Weights
            NumNeuronsPerLayer = new List<int> {numNeurons};
            _isLayered = numNeurons == 1;
            _weightMatrix = weightIni > 100 ? 
                Matrix.RandomMatrix(numInputs + numNeurons, numInputs + numNeurons, 1.0, 0.5) : 
                Matrix.ConstantMatrix(numInputs + numNeurons, numInputs + numNeurons, weightIni);

            // Selfloops are disallowed is there no synaptic delays
            // an output spike could cause artificial recursion
            for (var i = 0; i < _weightMatrix.Cols; i++)
            {
                _weightMatrix[i, i] = 0;
            }

            //Input neuron or neuron to input neuron connections are disallowed
            for (var i = 0; i < numInputs; i++)
                for (var n = 0; n < _weightMatrix.Rows; n++)
                    _weightMatrix[n, i] = 0;

            AssignNeurons(numInputs, numNeurons, neuronType, neuronProperties);
        }

        /// <summary>
        /// Creates a Spiking Neuron Network Object with a Layered Topology
        /// </summary>
        /// <param name="numInputs">Number of inputs to the network</param>
        /// <param name="numNeuronsPerLayer">Number of neurons per layer</param>
        /// <param name="neuronType">Neuron Type, inherited from ISpikingNeuron</param>
        /// <param name="weightIni">The initial weight between neurons.  If weightIni is > 100, then random weights are used with uniform distribution between -1 and 1, with non-zero probability of 0.5</param>
        /// <param name="neuronProperties">Optional neuron properties to assign to each neuron</param>
        public SpikingNeuronNetwork(int numInputs, IList<int> numNeuronsPerLayer, Type neuronType, double weightIni = 1e6, Dictionary<string, string> neuronProperties = null)
        {
            NumNeuronsPerLayer = numNeuronsPerLayer;
            // Create a layered weight matrix
            var numNeurons = numNeuronsPerLayer.Sum();
            var totalNeurons = numInputs + numNeurons;
            _isLayered = true;
            _weightMatrix = Matrix.ZeroMatrix(totalNeurons, totalNeurons);

            const double probability = 0.5;
            const int dispersion = 1;
            // Input layer
            for (var j = 0; j < numInputs; j++)
            {
                for (var k = numInputs; k < numInputs + numNeuronsPerLayer[0]; k++)
                {
                    var newWeight = weightIni > 100 ? ((_random.NextDouble() <= probability) ? dispersion * 2 * (_random.NextDouble() - 0.5) : 0) : weightIni;
                    _weightMatrix[j, k] = newWeight;
                }
            }
            // Hidden and output layers
            for (var layer = 1; layer < numNeuronsPerLayer.Count; layer++)
            {
                for (var j = numInputs; j < numInputs + numNeuronsPerLayer.Take(layer).Sum(); j++)
                {
                    for (var k = numInputs + numNeuronsPerLayer.Take(layer).Sum();
                         k < numInputs + numNeuronsPerLayer.Take(layer + 1).Sum();
                         k++)
                    {
                        var newWeight = weightIni > 100
                                            ? ((_random.NextDouble() <= probability)
                                                   ? dispersion * 2 * (_random.NextDouble() - 0.5)
                                                   : 0)
                                            : weightIni;
                        _weightMatrix[j, k] = newWeight;
                    }
                }
            }

            // Selfloops are disallowed without synaptic delays because 
            // an output spike could cause artificial recursion
            for (var i = 0; i < _weightMatrix.Cols; i++)
            {
                _weightMatrix[i, i] = 0;
            }

            //Input neuron or neuron to input neuron connections are disallowed
            for (var i = 0; i < numInputs; i++)
                for (var n = 0; n < _weightMatrix.Rows; n++)
                    _weightMatrix[n, i] = 0;

            AssignNeurons(numInputs, numNeurons, neuronType, neuronProperties);
        }

        /// <summary>
        /// Creates a new instance of Spiking Neuron Network for use in cloning
        /// </summary>
        /// <param name="random">Random number generator</param>
        /// <param name="weightMatrix">The weight matrix</param>
        /// <param name="isLayered">Indicates if the network is layered</param>
        /// <param name="numNeuronsPerLayer">The number of neurons per layer</param>
        /// <param name="neurons">The neuron dictionary, keyed by neuron index</param>
        private SpikingNeuronNetwork(Random random, Matrix weightMatrix, bool isLayered, IList<int> numNeuronsPerLayer, Dictionary<int, SpikingNeuron> neurons)
        {
            _random = random;
            _weightMatrix = weightMatrix;
            _isLayered = isLayered;
            _neurons = neurons;
            NumNeuronsPerLayer = numNeuronsPerLayer;
        }

        /// <summary>
        /// Creates a deep copy of the spiking neuron network
        /// </summary>
        /// <returns>A deep copy of the spiking neuron network</returns>
        public SpikingNeuronNetwork Clone()
        {
            var newNeurons = _neurons.ToDictionary(kvp => kvp.Key, kvp => (SpikingNeuron) kvp.Value.Clone());
            var clonedNetwork = new SpikingNeuronNetwork(_random, _weightMatrix, _isLayered, NumNeuronsPerLayer, newNeurons);
            foreach (var neuronIndices in newNeurons.Keys)
            {
                newNeurons[neuronIndices].NeuronNetwork = clonedNetwork;
            }
            return clonedNetwork;
        }

        /// <summary>
        /// Gets the stand alone neuron.
        /// </summary>
        /// <param name="neuronIndex">Index of the neuron.</param>
        /// <returns>The stand alone neuron.</returns>
        public ISpikingNeuron GetStandAloneNeuron(int neuronIndex)
        {
            return _neurons[neuronIndex].Clone();
        }

        /// <summary>
        /// Calculates Post Spike Derivatives for a specific neuron, that is the derivate of the state after the current input spike
        /// relative to the derivative of the state after the previous input spike
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of post spike derivatives</returns>
        public List<double> CalculatePostSpikeDerivatives(NeuronFiringHistory neuronFiringHistory)
        {
            return _neurons[neuronFiringHistory.NeuronIndex].CalculatePostSpikeDerivatives(neuronFiringHistory);
        }

        /// <summary>
        /// Calculates Post Spike Derivatives numerically for a specific neuron, that is the derivate of the state after the current input spike
        /// relative to the derivative of the state after the previous input spike
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of post spike derivatives</returns>
        public List<double> CalculatePostSpikeDerivativeNumerical(NeuronFiringHistory neuronFiringHistory)
        {
            return _neurons[neuronFiringHistory.NeuronIndex].CalculatePostSpikeDerivativesNumerical(neuronFiringHistory);
        }

        /// <summary>
        /// Calculates Output Spike Derivatives for a specific neuron, that is the derivate of the output spike time relative to both the
        /// weight on each synapse and the input spike time on each synapse
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of output spike time derivatives</returns>
        public Dictionary<Synapse, NeuronDerivativeParameters> CalculateOutputSpikeTimeDerivatives(NeuronFiringHistory neuronFiringHistory)
        {
            return _neurons[neuronFiringHistory.NeuronIndex].CalculateOutputSpikeTimeDerivatives(neuronFiringHistory);
        }

        /// <summary>
        /// Calculates Output Spike Derivatives numerically for a specific neuron, that is the derivate of the output spike time relative to both the
        /// weight on each synapse and the input spike time on each synapse
        /// </summary>
        /// <param name="neuronFiringHistory">The neuron firing history, includes a summary of states before and after all input spikes</param>
        /// <returns>A list of output spike time derivatives</returns>
        public Dictionary<Synapse, NeuronDerivativeParameters> CalculateOutputSpikeTimeDerivativesNumerical(NeuronFiringHistory neuronFiringHistory)
        {
            return _neurons[neuronFiringHistory.NeuronIndex].CalculateOutputSpikeTimeDerivativesNumerical(neuronFiringHistory);
        }

        /// <summary>
        /// Runs the spiking neuron network, processing the effect of all input spikes
        /// </summary>
        /// <param name="inputSpikes">List of input spikes to the network</param>
        /// <returns>Details of all generated spikes, both internal and output spikes</returns>
        public Dictionary<int, NeuronFiringHistory> RunSpikingNeuronNetwork(List<Spike> inputSpikes)
        {
            // Create a priority queue of NextSpikeTimes for each Neuron
            double currentTime = 0;
            const double endTime = 500;
            var neuronFiringHistories = new Dictionary<int, NeuronFiringHistory>();
            var spikeQueue = new SpikePriorityQueue();
            spikeQueue.AddRange(inputSpikes);
            foreach (var neuron in _neurons)
            {
                spikeQueue.Add(new Spike { Time = neuron.Value.NextSpikeTime, NeuronIndex = neuron.Value.NeuronIndex});
            }

            // Repeat until steady-state or for a fixed time
            while (currentTime < endTime && spikeQueue.Count > 0)
            {
                // Dequeue the next spike to process, eminating from Neuron A and add it to the outputspike List
                var currentSpike = spikeQueue.Dequeue();
                currentTime = currentSpike.Time;

                // Update Neuron A's State and add a new NextSpikeTime to the queue for Neuron A
                //     but skip this step if A is an input neuron
                if (currentSpike.NeuronIndex > NumInputs - 1)
                {
                    if (neuronFiringHistories.ContainsKey(currentSpike.NeuronIndex))
                    {
                        neuronFiringHistories[currentSpike.NeuronIndex].OutputSpikes.Add(currentSpike);
                    }
                    else
                    {
                        neuronFiringHistories.Add(currentSpike.NeuronIndex, new NeuronFiringHistory
                            {
                                NeuronIndex = currentSpike.NeuronIndex,
                                OutputSpikes = new List<Spike> { currentSpike },
                                SpikeStats = new SortedSet<SpikeStats>()
                            });
                    }                    
                    _neurons[currentSpike.NeuronIndex].State.Time = currentSpike.Time;
                    _neurons[currentSpike.NeuronIndex].State.StateVariable = _neurons[currentSpike.NeuronIndex].PostSpikeState;
                }
                
                // Get the set of neurons connected to Neuron A's output (Neurons B)
                var outputNeurons = GetNeuronsOutputNeurons(currentSpike.NeuronIndex);
                
                // Foreach neuron in B, update the State and update the NextSpikeTime in the queue
                foreach (var outputNeuron in outputNeurons)
                {
                    var oldSpike = new Spike {Time = outputNeuron.NextSpikeTime, NeuronIndex = outputNeuron.NeuronIndex};
                    SpikeStats currentSpikeStats;
                    var outputSpikeList = outputNeuron.ProcessSpike(currentSpike, out currentSpikeStats);
                    if (neuronFiringHistories.ContainsKey(outputNeuron.NeuronIndex))
                    {
                        neuronFiringHistories[outputNeuron.NeuronIndex].OutputSpikes.AddRange(outputSpikeList);
                        neuronFiringHistories[outputNeuron.NeuronIndex].SpikeStats.Add(currentSpikeStats);
                    }
                    else
                    {
                        neuronFiringHistories.Add(outputNeuron.NeuronIndex, new NeuronFiringHistory
                        {
                            NeuronIndex = outputNeuron.NeuronIndex,
                            OutputSpikes = new List<Spike>(outputSpikeList),
                            SpikeStats = currentSpikeStats == null ? new SortedSet<SpikeStats>() : new SortedSet<SpikeStats> {currentSpikeStats}
                        });
                    } 

                    // Find the outputNeuron in the spikeQueue and update the nextSpikeTime
                    spikeQueue.Update(oldSpike, outputNeuron.NextSpikeTime);
                }
            }

            return neuronFiringHistories;
        }

        /// <summary>
        /// Resets the network to its initial state
        /// </summary>
        public void ResetNetwork()
        {
            foreach (var neuron in _neurons)
            {
                neuron.Value.ResetState();
            }
        }

        /// <summary>
        /// Get a list of input synapses for a specific neuron
        /// </summary>
        /// <param name="neuronIndex">The neuron index</param>
        /// <returns>A list of input synapses for a specific neuron</returns>
        public List<Synapse> GetNeuronsInputSynapses(int neuronIndex)
        {
            return _neurons[neuronIndex].Weights.Select(x => x.Key).ToList();
        }

        /// <summary>
        /// Get a list of output synapses for a specific neuron
        /// </summary>
        /// <param name="neuronIndex">The neuron index</param>
        /// <returns>A list of output synapses for a specific neuron</returns>
        public List<Synapse> GetNeuronsOutputSynapses(int neuronIndex)
        {
            var weights = _weightMatrix.GetRow(neuronIndex);
            var outputNeuronIndices = weights.FindAllIndexOfNot(0);
            return outputNeuronIndices
                .Select(outputNeuronIndex => new Synapse(neuronIndex, outputNeuronIndex))
                .ToList();
        }

        /// <summary>
        /// Gets a list of neuron indices by layer
        /// </summary>
        /// <param name="layerIndex">The layer index</param>
        /// <returns>A list of neuron indices by layer</returns>
        public List<int> GetNeuronIndicesByLayer(int layerIndex)
        {
            if (!IsLayered || layerIndex >= NumNeuronsPerLayer.Count)
            {
                return null;
            }
            var startIndex = NumInputs + NumNeuronsPerLayer.Where((x,i) => i < layerIndex).Sum();
            return Enumerable.Range(startIndex, NumNeuronsPerLayer[layerIndex]).ToList();
        }

        /// <summary>
        /// Gets a list of output layer neuron indices
        /// </summary>
        /// <returns>A list of output layer neuron indices</returns>
        public List<int> GetOutputLayerNeuronIndices()
        {
            return GetNeuronIndicesByLayer(NumNeuronsPerLayer.Count - 1);
        }
        
        /// <summary>
        /// Gets the synaptic weight for a specific synapse
        /// </summary>
        /// <param name="synapse">The synapse</param>
        /// <returns>The synaptic weight for a specific synapse</returns>
        internal double GetSynapticWeight(Synapse synapse)
        {
            if (synapse.InputNeuronIndex > _weightMatrix.Rows) throw new ArgumentException("Invalid Synapse: " + synapse);
            if (synapse.OutputNeuronIndex > _weightMatrix.Cols) throw new ArgumentException("Invalid Synapse: " + synapse);
            return _weightMatrix[synapse.InputNeuronIndex, synapse.OutputNeuronIndex];
        }
        
        /// <summary>
        /// Gets the synaptic weights associated with a neuron's inputs
        /// </summary>
        /// <param name="neuronIndex">The neuron index</param>
        /// <returns>The synaptic weights associated with a neuron's inputs</returns>
        internal Dictionary<Synapse, double> GetNeuronWeights(int neuronIndex)
        {
            if (neuronIndex > _weightMatrix.Cols) return null;
            var weights = new Dictionary<Synapse, double>();
            var col = _weightMatrix.GetCol(neuronIndex); // Want Column neuronIndex for connections TO neuron
            for (var j = 0; j < col.Rows; j++)
            {
                if (Math.Abs(col[j, 0]) > Epsilon)
                {
                    weights.Add(new Synapse(j, neuronIndex), col[j, 0]);
                }
            }
            return weights;
        }

        /// <summary>
        /// Sets the synaptic weight for a specific synapse
        /// </summary>
        /// <param name="synapse">The synapse</param>
        /// <param name="newValue">The new synaptic weight</param>
        internal void SetSynapticWeight(Synapse synapse, double newValue)
        {
            if (synapse.InputNeuronIndex > _weightMatrix.Rows || synapse.OutputNeuronIndex > _weightMatrix.Cols) return;
            _weightMatrix[synapse.InputNeuronIndex, synapse.OutputNeuronIndex] = newValue;
        }

        /// <summary>
        /// Gets a list of a neuron's output neurons
        /// </summary>
        /// <param name="neuronIndex">The neuron index</param>
        /// <returns>A list of a neuron's output neurons</returns>
        internal List<SpikingNeuron> GetNeuronsOutputNeurons(int neuronIndex)
        {
            var weights = _weightMatrix.GetRow(neuronIndex);
            var outputNeuronIndices = weights.FindAllIndexOfNot(0);
            return outputNeuronIndices.Where(_neurons.ContainsKey)
                     .Select(x => _neurons[x])
                     .ToList();
        }

        /// <summary>
        /// Assigns neurons, all of type neuronType, to the network with optional neuronProperties
        /// </summary>
        /// <param name="numInputs">The number of inputs to the network</param>
        /// <param name="numNeurons">The number of neurons in the network</param>
        /// <param name="neuronType">The neuron type to populate the network with</param>
        /// <param name="neuronProperties">Optional dictionary of neuron properties</param>
        private void AssignNeurons(int numInputs, int numNeurons, Type neuronType, Dictionary<string,string> neuronProperties = null)
        {
            _neurons = new Dictionary<int, SpikingNeuron>();
            if (!typeof(ISpikingNeuron).IsAssignableFrom(neuronType))
            {
                throw new ArgumentException("Neuron type " + neuronType.Name + " must derive from ISpikingNeuron");
            }

            var neuronConstructor = neuronType.GetConstructor(new Type[] { });
            if (neuronConstructor == null)
            {
                throw new ArgumentException("Neuron type " + neuronType.Name + " must implement an empty constructor");
            }
            foreach (var neuronIndex in Enumerable.Range(numInputs, numNeurons))
            {
                _neurons.Add(neuronIndex, (SpikingNeuron)neuronConstructor.Invoke(new object[] { }));
                _neurons[neuronIndex].NeuronNetwork = this;
                _neurons[neuronIndex].NeuronIndex = neuronIndex;
                var typeProperties = neuronType.GetProperties();
                foreach (var neuronProperty in neuronProperties ?? new Dictionary<string, string>())
                {
                    var typeProperty = typeProperties.First(x => x.Name == neuronProperty.Key);
                    if (typeProperty == null || !typeProperty.CanWrite) continue;
                    var propType = typeProperty.PropertyType;
                    object castedValue;
                    if (propType.BaseType != null && propType.BaseType.Name == "Enum")
                    {
                        castedValue = Enum.Parse(propType, neuronProperty.Value);
                    }
                    else
                    {
                        castedValue = Convert.ChangeType(neuronProperty.Value, propType);                        
                    }

                    typeProperty.SetValue(_neurons[neuronIndex], castedValue);
                }
                _neurons[neuronIndex].ResetState();
            }
        }
    }
}
