﻿namespace SpikingNeuronNetwork.Lib
{
    using Interfaces;
    using System;
    using System.Collections.Generic;
    using System.Linq;
    
    /// <summary>
    /// Sum of Squared Spiking Error Class
    /// </summary>
    public class SumOfSquaredSpikingError : ISpikingError
    {
        /// <summary>
        /// Simulation Method Enum
        /// </summary>
        public SimulationMethod SimulationMethod { get; set; }

        /// <summary>
        /// Get the sum of squared spiking error between the desired and actual output spike trains
        /// </summary>
        /// <param name="desiredOutputSpikes">The desired output spike train</param>
        /// <param name="actualOutputSpikes">The actual output spike train</param>
        /// <returns>The error</returns>
        public double GetError(List<Spike> desiredOutputSpikes, List<Spike> actualOutputSpikes)
        {            
            if (desiredOutputSpikes.Count > actualOutputSpikes.Count)
            {
                throw new ArgumentException("Desired Spikes Count Must Not Be Greater Than The Actual Spike Count To Calculate Error");
            }

            if (actualOutputSpikes.Count > desiredOutputSpikes.Count)
            {
                actualOutputSpikes = actualOutputSpikes.Take(desiredOutputSpikes.Count).ToList();
            }

            desiredOutputSpikes = desiredOutputSpikes.OrderBy(x => x.Time).ToList();
            desiredOutputSpikes = desiredOutputSpikes.OrderBy(x => x.NeuronIndex).ToList();
            actualOutputSpikes = actualOutputSpikes.OrderBy(x => x.Time).ToList();
            actualOutputSpikes = actualOutputSpikes.OrderBy(x => x.NeuronIndex).ToList();

            double error = 0;
            for (var j = 0; j < desiredOutputSpikes.Count; j++)
            {
                if (desiredOutputSpikes[j].NeuronIndex != actualOutputSpikes[j].NeuronIndex)
                {
                    throw new ArgumentException("Desired and Actual Spikes Counts On Each Neuron Must Be Equal To Calculate Error");
                }
                error += Math.Pow(desiredOutputSpikes[j].Time - actualOutputSpikes[j].Time, 2);
            }
            return error/2;
        }

        /// <summary>
        /// Get the error to output spike time derivatives
        /// </summary>
        /// <param name="desiredOutputSpikes">The desired output spike train</param>
        /// <param name="actualOutputSpikes">The actual output spike train</param>
        /// <returns>A list of the error derivatives</returns>
        public List<double> GetErrorToOutputSpikeTimeDerivative(List<Spike> desiredOutputSpikes, List<Spike> actualOutputSpikes)
        {
            if (desiredOutputSpikes.Count > actualOutputSpikes.Count)
            {
                throw new ArgumentException("Desired Spike Count Must Not Be Greater Than the Actual Spike Count To Calculate Error");
            }

            if (actualOutputSpikes.Count > desiredOutputSpikes.Count)
            {
                actualOutputSpikes = actualOutputSpikes.Take(desiredOutputSpikes.Count).ToList();
            }

            var errorDerivatives = new List<double>();
            if (SimulationMethod == SimulationMethod.Numerical)
            {
                const double offset = 1e-4;
                foreach (var actualOutputSpike in actualOutputSpikes)
                {
                    actualOutputSpike.Time -= offset;
                    var error1 = GetError(desiredOutputSpikes, actualOutputSpikes);
                    actualOutputSpike.Time += 2*offset;
                    var error2 = GetError(desiredOutputSpikes, actualOutputSpikes);
                    actualOutputSpike.Time -= offset;
                    errorDerivatives.Add((error2-error1)/(2*offset));
                }
            }
            else
            {
                desiredOutputSpikes = desiredOutputSpikes.OrderBy(x => x.Time).ToList();
                desiredOutputSpikes = desiredOutputSpikes.OrderBy(x => x.NeuronIndex).ToList();
                actualOutputSpikes = actualOutputSpikes.OrderBy(x => x.Time).ToList();
                actualOutputSpikes = actualOutputSpikes.OrderBy(x => x.NeuronIndex).ToList();

                for (var j = 0; j < desiredOutputSpikes.Count; j++)
                {
                    if (desiredOutputSpikes[j].NeuronIndex != actualOutputSpikes[j].NeuronIndex)
                    {
                        throw new ArgumentException("Desired and Actual Spikes Counts On Each Neuron Must Be Equal To Calculate Error");
                    }
                    errorDerivatives.Add(actualOutputSpikes[j].Time - desiredOutputSpikes[j].Time);
                }
            }

            return errorDerivatives;                
        }
    }
}
