﻿namespace SpikingNeuronNetwork.Lib.Training
{
    using Interfaces;
    using System;
    using System.Collections.Generic;
    using System.Linq;

    /// <summary>
    /// Error Derivative Parameters Batch Class For Tracking Running Totals in Error Derivatives
    /// </summary>
    public class ErrorDerivativeParametersBatch : IErrorDerivativeParameters
    {
        /// <summary>
        /// Gets or sets the Error Derivative Total Dictionary between synapses and the running total of error derivatives and the count of the number error derivatives added
        /// </summary>
        public Dictionary<Synapse, Tuple<double, int>> ErrorDerivativeTotal { get; set; }

        /// <summary>
        /// Creates a new instance of ErrorDerivativeParametersBatch from an ErrorDerivateParameters object
        /// </summary>
        /// <param name="errorDerivativeParameters">The ErrorDerivativeParameters object</param>
        public ErrorDerivativeParametersBatch(ErrorDerivativeParameters errorDerivativeParameters)
        {
            ErrorDerivativeTotal = new Dictionary<Synapse, Tuple<double, int>>();
            foreach (var synapse in errorDerivativeParameters.GetSynapses())
            {
                ErrorDerivativeTotal.Add(synapse, new Tuple<double, int>(errorDerivativeParameters.GetErrorDerivative(synapse), 1));
            }
        }

        /// <summary>
        /// Adds errorDerivativeValue to the ErrorDerivativeTotal dictionary at synapse
        /// </summary>
        /// <param name="synapse">The synapse</param>
        /// <param name="errorDerivativeValue">The error derivative value</param>
        public void AddErrorDerivative(Synapse synapse, double errorDerivativeValue)
        {
            if (ErrorDerivativeTotal.ContainsKey(synapse))
            {
                ErrorDerivativeTotal[synapse] = new Tuple<double, int>(
                    ErrorDerivativeTotal[synapse].Item1 + errorDerivativeValue,
                    ErrorDerivativeTotal[synapse].Item2 + 1);
            }
            else
            {
                ErrorDerivativeTotal.Add(synapse, new Tuple<double, int>(errorDerivativeValue, 1));
            }            
        }

        /// <summary>
        /// Gets a list of the synapses for which there are error derivative parameters
        /// </summary>
        /// <returns>The list of the synapses for which there are error derivative parameters</returns>
        public List<Synapse> GetSynapses()
        {
            return ErrorDerivativeTotal.Keys.ToList();
        }

        /// <summary>
        /// Get the derivate of the error relative change in weight for a certain synapse
        /// </summary>
        /// <param name="synapse">The synapse</param>
        /// <returns>The error derivative</returns>
        public double GetErrorDerivative(Synapse synapse)
        {
            return ErrorDerivativeTotal[synapse].Item1 / ErrorDerivativeTotal[synapse].Item2;
        }
    }
}
