﻿namespace SpikingNeuronNetwork.Lib.Training
{
    using Interfaces;
    using System;
    using System.Collections.Generic;
    using System.Linq;

    /// <summary>
    /// Train Network Class
    /// </summary>
    public class TrainNetwork
    {
        /// <summary>
        /// Gets or sets the training statistics
        /// </summary>
        public TrainingStats TrainingStats { get; set; }

        /// <summary>
        /// Creates a new instance the TrainNetwork class
        /// </summary>
        /// <param name="trainingStats">The initial training statistics parameters</param>
        public TrainNetwork(TrainingStats trainingStats)
        {
            TrainingStats = trainingStats;
        }

        /// <summary>
        /// Creates a new instance the TrainNetwork class
        /// </summary>
        /// <param name="spikingNeuronNetwork">The spiking neuron network</param>
        /// <param name="trainingSet">The training set</param>
        /// <param name="learningRate">The learning rate</param>
        /// <param name="maximumErrorAfterTraining">The maximum error after training</param>
        /// <param name="maximumNumberOfTrainingEpochs">The maximum number of training epochs</param>
        /// <param name="errorCalculator">The error calculator</param>
        /// <param name="trainingAlgorithm">The training algorithm</param>
        /// <param name="trainingMethod">The training method</param>
        public TrainNetwork(SpikingNeuronNetwork spikingNeuronNetwork, List<SpikeSet> trainingSet, double learningRate, double maximumErrorAfterTraining, double maximumNumberOfTrainingEpochs, ISpikingError errorCalculator, ITrainingAlgorithm trainingAlgorithm, TrainingMethod trainingMethod)
        {
            TrainingStats = new TrainingStats
                {
                    OriginalNetwork = spikingNeuronNetwork,
                    TrainingSet = trainingSet,
                    LearningRate = learningRate,
                    MaximumErrorAfterTraining = maximumErrorAfterTraining,
                    MaximumNumberOfTrainingEpochs = maximumNumberOfTrainingEpochs,
                    ErrorCalculator = errorCalculator,
                    TrainingMethod = trainingMethod,
                    TrainingAlgorithm = trainingAlgorithm
                };
            TrainingStats.ResetTraining();
        }

        /// <summary>
        /// Calculate the error derivative
        /// </summary>
        /// <param name="neuronFiringHistories">The history of firing for each neuron in the network</param>
        /// <param name="desiredOutputSpikeTimes">The desired output spike times</param>
        /// <returns>The error derivative parameters</returns>
        public ErrorDerivativeParameters CalculateErrorDerivative(Dictionary<int, NeuronFiringHistory> neuronFiringHistories, List<Spike> desiredOutputSpikeTimes)
        {
            var neuronDerivativeParameters = new Dictionary<Synapse, NeuronDerivativeParameters>();
            foreach (var value in neuronFiringHistories.Values)
                neuronDerivativeParameters = neuronDerivativeParameters.Concat(TrainingStats.CurrentNetwork.CalculateOutputSpikeTimeDerivatives(value)).ToDictionary(x => x.Key, x => x.Value);
            TrainingStats.ErrorCalculator.SimulationMethod = SimulationMethod.EventDriven;

            return CalculateErrorDerivativeHelper(neuronFiringHistories, desiredOutputSpikeTimes,
                                                  neuronDerivativeParameters);
        }

        /// <summary>
        /// Calculate the error derivative numerically
        /// </summary>
        /// <param name="neuronFiringHistories">The history of firing for each neuron in the network</param>
        /// <param name="desiredOutputSpikeTimes">The desired output spike times</param>
        /// <returns>The error derivative parameters</returns>
        public ErrorDerivativeParameters CalculateErrorDerivativeNumerical(Dictionary<int, NeuronFiringHistory> neuronFiringHistories, List<Spike> desiredOutputSpikeTimes)
        {
            var neuronDerivativeParameters = new Dictionary<Synapse, NeuronDerivativeParameters>();
            foreach (var value in neuronFiringHistories.Values)
                neuronDerivativeParameters = neuronDerivativeParameters.Concat(TrainingStats.CurrentNetwork.CalculateOutputSpikeTimeDerivativesNumerical(value)).ToDictionary(x => x.Key, x => x.Value);
            TrainingStats.ErrorCalculator.SimulationMethod = SimulationMethod.Numerical;

            return CalculateErrorDerivativeHelper(neuronFiringHistories, desiredOutputSpikeTimes,
                                                  neuronDerivativeParameters);
        }

        /// <summary>
        /// Start training the network
        /// </summary>
        /// <returns>Indicates if training was successful</returns>
        public bool StartTraining()
        {
            if (TrainingStats.CurrentNetwork == null)
            {
                TrainingStats.ResetTraining();
            }

            if (!TrainingStats.CurrentNetwork.IsLayered)
            {
                return false;
            }

            ErrorDerivativeParametersBatch previousErrorDerivativesBatch = null;
            int epoch;
            for (epoch = 1; epoch < TrainingStats.MaximumNumberOfTrainingEpochs; epoch++)
            {
                Console.WriteLine("Epoch: " + epoch);
                var trainingsetNumber = 0;
                var currentError = 0.0;
                var nonFiringNeuronIndices = new List<int>();
                ErrorDerivativeParametersBatch errorDerivativesBatch = null;
                ErrorDerivativeParameters previousErrorDerivatives = null;                
                foreach (var trainingPattern in TrainingStats.TrainingSet)
                {
                    TrainingStats.CurrentNetwork.ResetNetwork();
                    var neuronFiringHistories = TrainingStats.CurrentNetwork.RunSpikingNeuronNetwork(trainingPattern.InputSpikes);
                    nonFiringNeuronIndices.AddRange(NeuronFiringHistory.GetNonFiringNeuronIndices(neuronFiringHistories));
                    var errorDerivatives = CalculateErrorDerivative(neuronFiringHistories, trainingPattern.OutputSpikes);

                    foreach (var outputNeuronIndex in TrainingStats.CurrentNetwork.GetOutputLayerNeuronIndices())
                    {
                        var currentOutputSpikes = trainingPattern.OutputSpikes.Where(x => x.NeuronIndex == outputNeuronIndex).ToList();
                        currentError += TrainingStats.ErrorCalculator.GetError(currentOutputSpikes, neuronFiringHistories[outputNeuronIndex].OutputSpikes);
                    }

                    if (TrainingStats.TrainingMethod == TrainingMethod.Online)
                    {
                        TrainingStats.TrainingAlgorithm.UpdateWeights(TrainingStats, errorDerivatives, previousErrorDerivatives);
                    }
                    else if (errorDerivatives != null)
                    {
                        if (errorDerivativesBatch == null)
                        {
                            errorDerivativesBatch = new ErrorDerivativeParametersBatch(errorDerivatives);
                        }
                        else
                        {
                            foreach (var synapse in errorDerivatives.GetSynapses())
                            {
                                errorDerivativesBatch.AddErrorDerivative(synapse, errorDerivatives.GetErrorDerivative(synapse));
                            }
                        }
                    }

                    previousErrorDerivatives = errorDerivatives;
                    trainingsetNumber++;
                }
                if (TrainingStats.TrainingMethod == TrainingMethod.Batch)
                {
                    TrainingStats.TrainingAlgorithm.UpdateWeights(TrainingStats, errorDerivativesBatch, previousErrorDerivativesBatch);
                    previousErrorDerivativesBatch = errorDerivativesBatch;
                }
                nonFiringNeuronIndices = nonFiringNeuronIndices.Distinct().OrderBy(x => x).ToList();
                TrainingStats.PerEpochStats.Add(TrainingStats.NumEpochs + epoch, new TrainingStatsPerEpoch(currentError, nonFiringNeuronIndices));
                Console.WriteLine("\tError on training set " + trainingsetNumber + ": " + currentError);
                if (currentError >= TrainingStats.MaximumErrorAfterTraining) continue;
                TrainingStats.NumEpochs += epoch;
                return true;
            }
            TrainingStats.NumEpochs += epoch;
            return false;
        }

        #region Helper Methods

        /// <summary>
        /// Helper method to calculate the network error derivatives during training
        /// </summary>
        /// <param name="neuronFiringHistories">The neuron firing histories</param>
        /// <param name="desiredOutputSpikeTimes">The desired output spike times</param>
        /// <param name="neuronDerivativeParameters">Neuron derivative parameters on a per synapse basis</param>
        /// <returns>The network error derivative parameters</returns>
        private ErrorDerivativeParameters CalculateErrorDerivativeHelper(IReadOnlyDictionary<int, NeuronFiringHistory> neuronFiringHistories, List<Spike> desiredOutputSpikeTimes, IReadOnlyDictionary<Synapse, NeuronDerivativeParameters> neuronDerivativeParameters)
        {
            var errorDerivativeParameters = new ErrorDerivativeParameters();

            #region Output Neurons
            foreach (var outputNeuronIndex in TrainingStats.CurrentNetwork.GetOutputLayerNeuronIndices())
            {
                if (desiredOutputSpikeTimes.Count(x => x.NeuronIndex == outputNeuronIndex) > 1)
                {
                    throw new ArgumentException("MIMO Training Not Implemented");
                }

                if (neuronFiringHistories[outputNeuronIndex].OutputSpikes.Count == 0 || desiredOutputSpikeTimes.Count(x => x.NeuronIndex == outputNeuronIndex) == 0)
                {
                    throw new ArgumentException("Actual or Desired Output Spike Times are Missing");
                }

                errorDerivativeParameters.ErrorToOutputSpikeTimeDerivatives.Add(
                    outputNeuronIndex,
                    TrainingStats.ErrorCalculator.GetErrorToOutputSpikeTimeDerivative(
                        desiredOutputSpikeTimes.Where(x => x.NeuronIndex == outputNeuronIndex).ToList(),
                        neuronFiringHistories[outputNeuronIndex].OutputSpikes).First());

                // Where clause here is because not all input synapses may have spikes
                foreach (var synapse in TrainingStats.CurrentNetwork.GetNeuronsInputSynapses(outputNeuronIndex).Where(neuronDerivativeParameters.ContainsKey))
                {
                    errorDerivativeParameters.OutputSpikeTimeToWeightDerivatives.Add(synapse, neuronDerivativeParameters[synapse].OutputSpikeTimeToWeightDerivative);
                }
            }
            #endregion

            #region Hidden Neurons
            foreach (var layerIndex in Enumerable.Range(0, TrainingStats.CurrentNetwork.NumNeuronsPerLayer.Count - 1).Reverse()) // - 1 because output layer is already done
            {
                foreach (var hiddenNeuronIndex in TrainingStats.CurrentNetwork.GetNeuronIndicesByLayer(layerIndex))
                {
                    var outputSynapses = TrainingStats.CurrentNetwork.GetNeuronsOutputSynapses(hiddenNeuronIndex);
                    errorDerivativeParameters.ErrorToOutputSpikeTimeDerivatives.Add(hiddenNeuronIndex, 0);
                    foreach (var outputSynapse in outputSynapses)
                    {
                        // ToDo: Confirm that OutputSpikeTimeToInputSpikeTimeDerivative reduces to 0 if 
                        // ToDo: the output neuron's output spike happens before the input neuron's output spike
                        errorDerivativeParameters.ErrorToOutputSpikeTimeDerivatives[hiddenNeuronIndex] +=
                            errorDerivativeParameters.ErrorToOutputSpikeTimeDerivatives[outputSynapse.OutputNeuronIndex] *
                            neuronDerivativeParameters[outputSynapse].OutputSpikeTimeToInputSpikeTimeDerivative;
                    }
                    foreach (var synapse in TrainingStats.CurrentNetwork.GetNeuronsInputSynapses(hiddenNeuronIndex))
                    {
                        errorDerivativeParameters.OutputSpikeTimeToWeightDerivatives.Add(synapse, neuronDerivativeParameters[synapse].OutputSpikeTimeToWeightDerivative);
                    }
                }
            }
            #endregion

            return errorDerivativeParameters;
        }

        #endregion
    }
}
