﻿namespace SpikingNeuronNetwork.Test
{
    using Lib;
    using Lib.NeuronModels;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Linq;
    
    /// <summary>
    /// Tests ability to run a spiking neuron network
    /// </summary>
    [TestClass]
    public class SpikingNeuronNetworkRunningTests
    {
        /// <summary>
        /// Tests ability to setup a layered spiking neuron network
        /// </summary>
        [TestMethod]
        public void ThetaNeuronNetworkLayeredTest()
        {
            const int numInputs = 3;
            var numNeuronsPerLayer = new[] { 1, 2 };
            var network = new SpikingNeuronNetwork(numInputs, numNeuronsPerLayer, typeof(ThetaNeuron), 1);
            Assert.IsTrue(network.IsLayered);
            Assert.IsTrue(new List<int> { 3 }.SequenceEqual(network.GetNeuronIndicesByLayer(0)));
            Assert.IsTrue(new List<int> { 4, 5 }.SequenceEqual(network.GetNeuronIndicesByLayer(1)));
            network = new SpikingNeuronNetwork(numInputs, numNeuronsPerLayer.Sum(), typeof(ThetaNeuron), 1);
            Assert.IsFalse(network.IsLayered);
        }

        /// <summary>
        /// Tests ability to run a Spiking Neuron Network containing a single neuron
        /// and one input
        /// </summary>
        [TestMethod]
        public void ThetaNeuronNetworkBasicRunCompareOneInputTest()
        {
            const int numInputs = 1;
            const int numNeurons = 1;
            const double weightIni = 0.15;
            const double baselineCurrent = 0.001;
            var inputSpikeTimes = new List<Spike>
                {
                    new Spike { Time = 1, NeuronIndex = 0}, 
                };

            SpikeTrainTestHelper(numInputs, numNeurons, weightIni, baselineCurrent, inputSpikeTimes);
        }

        /// <summary>
        /// Tests ability to run a Spiking Neuron Network containing a single neuron
        /// and two inputs
        /// </summary>
        [TestMethod]
        public void ThetaNeuronNetworkBasicRunCompareTwoInputsTest()
        {
            const int numInputs = 2;
            const int numNeurons = 1;
            const double weightIni = 0.1;
            const double baselineCurrent = 0.001;
            var inputSpikeTimes = new List<Spike>
                {
                    new Spike { Time = 2, NeuronIndex = 0}, 
                    new Spike { Time = 1, NeuronIndex = 1}, 
                };

            SpikeTrainTestHelper(numInputs, numNeurons, weightIni, baselineCurrent, inputSpikeTimes);
        }

        /// <summary>
        /// Tests ability to run a Spiking Neuron Network containing a single neuron
        /// and three inputs
        /// </summary>
        [TestMethod]
        public void ThetaNeuronNetworkBasicRunCompareThreeInputsTest()
        {
            const int numInputs = 3;
            const int numNeurons = 1;
            const double weightIni = 0.01;
            const double baselineCurrent = 0.001;
            var inputSpikeTimes = new List<Spike>
                {
                    new Spike { Time = 2, NeuronIndex = 0}, 
                    new Spike { Time = 1, NeuronIndex = 1}, 
                    new Spike { Time = 2.1, NeuronIndex = 2}
                };

            SpikeTrainTestHelper(numInputs, numNeurons, weightIni, baselineCurrent, inputSpikeTimes);
        }

        /// <summary>
        /// Tests ability to run a Spiking Neuron Network containing a single neuron
        /// and one input which produces no output spikes
        /// </summary>
        [TestMethod]
        public void ThetaNeuronNetworkBasicRunNoOutputSpikesTest()
        {
            const int numInputs = 1;
            const int numNeurons = 1;
            const double weightIni = -0.15;
            const double baselineCurrent = -0.001;
            var inputSpikeTimes = new List<Spike>
                {
                    new Spike { Time = 0.01, NeuronIndex = 0}, 
                };

            SpikeTrainTestHelper(numInputs, numNeurons, weightIni, baselineCurrent, inputSpikeTimes);
        }

        /// <summary>
        /// Tests ability to run a Spiking Neuron Network containing a single neuron
        /// and one input that takes place before the network time, and is thus ignored
        /// </summary>
        [TestMethod]
        public void ThetaNeuronNetworkBasicRunAllInputSpikesComeBeforeNetworkTime()
        {
            const int numInputs = 1;
            const int numNeurons = 1;
            const double weightIni = -0.15;
            const double baselineCurrent = -0.001;
            const int currentNeuron = numInputs + numNeurons - 1;
            var inputSpikeTimes = new List<Spike>
                {
                    new Spike { Time = 0.01, NeuronIndex = 0}, 
                };

            var thetaNeuronNetwork = new SpikingNeuronNetwork(numInputs, numNeurons, typeof(ThetaNeuron), weightIni, new Dictionary<string, string> { {"Io", baselineCurrent.ToString(CultureInfo.InvariantCulture)}, {"Method", "EventDriven"} });
            thetaNeuronNetwork.RunSpikingNeuronNetwork(inputSpikeTimes);
            inputSpikeTimes.First().Time -= 0.005;
            var neuronFiringHistories = thetaNeuronNetwork.RunSpikingNeuronNetwork(inputSpikeTimes);

            Assert.AreEqual(1, neuronFiringHistories.Count);
            Assert.AreEqual(currentNeuron, neuronFiringHistories.First().Key);
            Assert.AreEqual(-1, neuronFiringHistories.First().Value.StartTime);
            Assert.AreEqual(-1, neuronFiringHistories.First().Value.EndTime);
            Assert.AreEqual(0, neuronFiringHistories.First().Value.OutputSpikes.Count);
            Assert.AreEqual(0, neuronFiringHistories.First().Value.SpikeStats.Count);
        }

        /// <summary>
        /// Tests that a spiking neuron network is not created with an invalid spiking neuron type
        /// </summary>
        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void ThetaNeuronNetworkBadNeuronTypeTest()
        {
            const int numInputs = 2;
            const int numNeurons = 1;
            const double weightIni = 0.1;
            // ReSharper disable ObjectCreationAsStatement
            new SpikingNeuronNetwork(numInputs, numNeurons, typeof(double), weightIni);
            // ReSharper restore ObjectCreationAsStatement
        }

        #region Helper Methods

        /// <summary>
        /// Helper method to run and test a spiking neuron network with a random topology.
        /// The network as a whole is run in an event-driven manner.  Then a single output
        /// neuron is run numerically.  The spike trains of the two runs are asserted to be
        /// equal within a small tolerance.
        /// </summary>
        /// <param name="numInputs">Number of inputs to the network</param>
        /// <param name="numNeurons">Number of neurons within the network</param>
        /// <param name="weightIni">The initial weight between neurons</param>
        /// <param name="baselineCurrent">The baseline current to set each of the network's neurons to</param>
        /// <param name="inputSpikeTimes">List of input spike times</param>
        private static void SpikeTrainTestHelper(int numInputs, int numNeurons, double weightIni, double baselineCurrent, List<Spike> inputSpikeTimes)
        {
            var currentNeuron = numInputs + numNeurons - 1;
            var thetaNeuronNetwork = new SpikingNeuronNetwork(numInputs, numNeurons, typeof(ThetaNeuron), weightIni, new Dictionary<string, string> { {"Io", baselineCurrent.ToString(CultureInfo.InvariantCulture)}, {"Method", "EventDriven"} });
            var neuronFiringHistories = thetaNeuronNetwork.RunSpikingNeuronNetwork(inputSpikeTimes);

            // Compare Running With Stand-Alone Neuron, numerically
            var outputSpikeTimesNetwork = neuronFiringHistories[currentNeuron].OutputSpikes;
            var thetaNeuron = (ThetaNeuron)thetaNeuronNetwork.GetStandAloneNeuron(currentNeuron); //new ThetaNeuron(numInputs, weightIni) {Method = ThetaNeuron.SimulationMethod.Numerical};
            thetaNeuron.Method = SimulationMethod.Numerical;
            thetaNeuron.ResetState();
            var outputSpikeTimesStandAlone = thetaNeuron.RunThetaNeuron(inputSpikeTimes).ToList();
            TestHelper.CompareSpikeTrains(outputSpikeTimesNetwork, outputSpikeTimesStandAlone, 0.01);
        }

        #endregion
    }
}
