﻿namespace SpikingNeuronNetwork.Test
{
    using Lib;
    using Lib.Interfaces;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using System;
    using System.Collections.Generic;
    using System.Linq;

    /// <summary>
    /// Set of helper methods for various spiking neuron network tests
    /// </summary>
    public class TestHelper
    {
        /// <summary>
        /// Small value is used when comparing floats in various tests
        /// </summary>
        public static double Epsilon = 1e-4;

        /// <summary>
        /// Finds the mean squared error in spike times between two spike trains.
        /// Asserts that the MSE is within threshold. Also, if the number of spikes 
        /// in each is different, then failure is asserted.
        /// </summary>
        /// <param name="spikeTrain1">The first spike train</param>
        /// <param name="spikeTrain2">The second spike train</param>
        /// <param name="threshold">The comparison threshold</param>
        public static void CompareSpikeTrains(IReadOnlyCollection<Spike> spikeTrain1, IReadOnlyList<Spike> spikeTrain2, double threshold)
        {
            //Find Mean-Squared in Spike Times Between the Different Methods
            Assert.AreEqual(spikeTrain1.Count, spikeTrain2.Count,
                            "The Two Spike Trains Contain Different Numbers of Spikes! Unable to Compare.\n\tTrain 1: " + String.Join(", ", spikeTrain1) + "\n\tTrain 2: " + String.Join(", ", spikeTrain2));

            if (spikeTrain1.Count == 0)
            {
                Console.WriteLine("The Two Spike Trains Produced No Output Spikes!");
                Console.WriteLine();
            }
            else
            {
                double meanSquaredError = 0;
                meanSquaredError +=
                    spikeTrain1.Select((t, j) => Math.Pow(t.Time - spikeTrain2[j].Time, 2)).Sum();
                meanSquaredError /= spikeTrain1.Count;
                Assert.IsTrue(meanSquaredError < threshold,
                              "Output Spike Mean-Squared Error Between The Two Spike Trains is greater than " + threshold);
                Console.WriteLine("Output Spike Mean-Squared Error Between The Two Spike Trains: ");
                Console.WriteLine("\t" + meanSquaredError);
                Console.WriteLine();
            }
        }

        /// <summary>
        /// For a single spiking neuron, compares the output spikes (up to maxNumOutputSpikes) produced by 
        /// running the network event-driven and numerically. The spike trains of the two runs are asserted to be
        /// equal within a small tolerance.
        /// </summary>
        /// <param name="neuron">The spike neuron</param>
        /// <param name="inputSpikeTimes">The input spikes</param>
        /// <param name="maxNumOutputSpikes">The maximum number of output spikes to compare</param>
        public static void CompareNumericalAndEventDrivenSpikeTrains(ISpikingNeuron neuron, List<Spike> inputSpikeTimes, int maxNumOutputSpikes = 3)
        {
            inputSpikeTimes.RemoveAll(x => x.NeuronIndex >= neuron.NumInputs);

            // Display Neuron Properties
            Console.WriteLine(neuron);

            // Simulate Theta Neuron Numerically
            neuron.Method = SimulationMethod.Numerical;
            var outputSpikeTimesNumerical = neuron.RunThetaNeuron(inputSpikeTimes, maxNumOutputSpikes).ToList();

            //Simulate Theta Neuron with Event-Driven Methodology
            neuron.ResetState();
            neuron.Method = SimulationMethod.EventDriven;
            var outputSpikeTimesEventDriven = neuron.RunThetaNeuron(inputSpikeTimes, maxNumOutputSpikes).ToList();
            CompareSpikeTrains(outputSpikeTimesNumerical, outputSpikeTimesEventDriven, 0.01);
        }
    }
}
